using System;
using System.Runtime.InteropServices;
using System.Collections.Generic;
using System.Text;
using System.Resources;
using System.Reflection;
using System.IO;

/// Feature search option (visibility) example (in JSON format). All fields are optional
/*
   {
    // Perform feature search only inside of specified license. By default, the feature search is performed in all licenses
    "licenseId": 0x12345678,
    // The mask that specifying what kind of features can be found - local (0x01), remote (0x02) or both (0x03). See GrdRemoteMode enumeration.
    // By default, the search is performed for all features
    "remoteMode": 3,
    // The mask for specifying the dongle models to which feature can be bound to (0 means all models). See GrdDongleModel enumeration.
    // By default, the feature search is performed in all dongle models
    "dongleModel": 0x400,
    // The number of the product to which feature bound to (0 means all products). By default, the feature search is performed in all products
    "productNumber": 777,
    // Number of network resource to consume. By default, 1
    "networkResourceToConsume" : 1,
    // Configuration for Guardant Control Center
    "controlCenter":
    {
        // HostName or IP address of the computer on which Guardant Control Center is running
       "hostName": [ "192.168.77.77" ],
       // Broadcast search for remote licenses (enabled by default)
       "broadcastSearch" : true,
       // Ipv4 broadcast address (255.255.255.255 by default)
       "broadcastAddress": "255.255.255.255"
    }
   }
*/

namespace grdlic
{
    public enum Status : int
    {
        OK                                          = 0,       // Successful operation status
        INVALID_HANDLE                              = 1,       // Invalid handle
        FEATURE_NOT_FOUND                           = 2,       // Feature not found
        FEATURE_RESOURCE_EXHAUST                    = 3,       // Feature network resource exhausted
        INVALID_FEATURE                             = 4,       // Invalid feature data
        INVALID_SERIAL_NUMBER                       = 5,       // Serial number is invalid
        INVALID_LICENSE                             = 6,       // Can't read DL license file because of hardware mismatched or part of file header is corrupted
        LICENSE_NOT_ACTIVATED                       = 7,       // License with this licenseId has not been activated yet or reactivation required
        LICENSE_ALREADY_ACTIVATED                   = 8,       // License already activated
        GET_HARDWARE_ID_ERROR                       = 9,       // Can not get hardware id
        NOT_ENOUGH_HARWARE_TO_BIND                  = 10,      // Insufficient number of computer components to activate
        REQUEST_FAILED                              = 12,      // Unable to send request
        GET_RESPONSE_FAILED                         = 13,      // Unable to receive response
        SERVER_NOT_FOUND                            = 14,      // Unable to find server
        INCORRECT_RESPONSE_DATA                     = 15,      // Incorrect response format or data not valid
        FILE_CREATION_ERROR                         = 16,      // Error during creation of file (e.g. license file)
        SERIAL_NUMBER_RESOURCE_EXHAUSTED            = 17,      // Serial number resource exhausted
        NO_UPDATES_AVAILABLE                        = 18,      // No updates available for license
        ACTIVATION_NOT_AVAILABLE                    = 19,      // Activation not enabled by software vendor reasons
        NO_CUSTOMER_SPECIFIED                       = 20,      // No customer info specified
        CAN_NOT_ACTIVATE_NON_DEMO_LICENSE           = 21,      // Can't activate non demo license
        OUTDATED_API_VERSION                        = 22,      // API version too low
        INVALID_PARAMETER                           = 23,      // Invalid function argument
        BUFFER_TOO_SMALL                            = 24,      // Buffer is not enough to perform operation
        INTERNAL_ERROR                              = 25,      // Internal error
        LICENSE_IS_BANNED                           = 26,      // Software vendor has banned serial number and license
        NO_LICENSE_AVAILABLE                        = 27,      // No license available (e.g. order access time is over)
        FEATURE_RUNCOUNTER_EXHAUST                  = 28,      // Feature execution counter is exhausted
        FEATURE_EXPIRED                             = 29,      // Feature life time is expired
        INVALID_PUBLIC_KEY                          = 30,      // Invalid public key
        NO_SERVICE                                  = 31,      // Service is not supported
        UNABLE_SEND_REQUEST_TO_ADMIN_RUNTIME        = 32,      // Admin runtime is not installed or outdated
        DONGLE_NOT_FOUND                            = 33,      // Guardant dongle not found
        NO_RESULTS_FOUND                            = 34,      // No any results found
        INVALID_DIGEST                              = 35,      // Invalid message or digest
        FEATURE_INACTIVE                            = 36,      // The period of using feature has not yet arrived or the feature is not active
        MEMORY_OUT_OF_RANGE                         = 37,      // Access to memory beyond its range
        ACCESS_DENIED                               = 38,      // Access denied
        NUMBER_ATTEMPTS_EXHAUSTED                   = 39,      // The number of password attempts has been exhausted
        SERIAL_NUMBER_NOT_FOUND                     = 40,      // Such serial number not found on server
        INVALID_MEMORY                              = 41,      // Invalid memory ID specified or item does not exist
        VIRTUAL_ENVIRONMENT_DETECTED                = 42,      // Running in virtual environment detected
        REMOTE_DESKTOP_DETECTED                     = 43,      // Detected running in remote desktop mode
        OUTDATED_GCC_VERSION                        = 44,      // Outdated Gurdant Control Center version
        REHOST_IS_NOT_ALLOWED                       = 45,      // Rehost is not allowed
        STORAGE_CORRUPTED                           = 46,      // DL storage corrupted
        STORAGE_ID_MISMATCH                         = 47,      // DL storage ID mismatch (input is not intended for current storage)
        INVALID_UPDATE_OR_ALREADY_INSTALLED         = 48,      // Invalid update for DL received or update has been installed before
        INVALID_REHOST_DATA                         = 49,      // Invalid rehost data sent to server
        HARDWARE_ID_MISMATCH                        = 50,      // Hardware id mismatch
        REACTIVATION_IS_NOT_ALLOWED                 = 51,      // Reactivation is not allowed because of license contains write memory or features with counter
        RECIPIENT_HAS_THE_SAME_LICENSE              = 52,      // Recipient has the same license activated as the license trying to rehost to it
        CAN_NOT_DETACH_SOME_FEATURES                = 53,      // Detach failed for some of chosen features
        CANCEL_PREVIOUS_DETACH_REQUIRED             = 54,      // Required cancel previous detach before making new one
        DETACHED_LICENSE_EXPIRED                    = 55,      // Detached license already exprired
        DETACHED_LICENSE_NOT_FOUND                  = 56,      // Detached license not found
        DETACH_FORBIDDEN_BY_GCC                     = 57,      // Detach forbidden by Gurdant Control Center
        TRIAL_LICENSE_EXPIRED                       = 58,      // Trial license expired
        OPERATION_IS_NOT_ALLOWED                    = 59,      // Operation is not allowed
        RTC_ERROR                                   = 60,      // Timer error
        FUNCTION_NOT_FOUND                          = 0x70000000, // Function not found
        DLL_NOT_FOUND                               = 0x70000001, // DLL not found
        MANAGE_ERROR                                = 0x70000002  // Unknown error
    }

    public enum HandleValue: uint {INVALID = 0xFFFFFFFF};

    public enum GrdECC160 : int
    {
        PUBLIC_KEY_SIZE = 40,
        PRIVATE_KEY_SIZE = 20,
        DIGEST_SIZE = 40,
        MESSAGE_SIZE = 20
    }

    [Flags]
    public enum RemoteMode : int
    {
        LOCAL    = 0x01,
        REMOTE   = 0x02
    }

    [Flags]
    public enum DongleModel : int
    {
        SIGN    = 0x80,
        CODE    = 0x100,
        DL      = 0x400
    }

    [Flags]
    public enum FeatureEncryptMode : int
    {
        GRD_EM_ECB                = 0x00,       // Electronic Code Book encryption mode (set by default). Data size must be multiple of 16
        GRD_EM_CBC                = 0x01,       // Cipher Block Chaining encryption mode. Data size must be multiple of 16
        GRD_EM_CFB                = 0x02,       // Cipher Feed Back encryption mode. Is allowed only if GRD_SOFTWARE_ACCELERATION flag is provided
        GRD_EM_OFB                = 0x04,       // Output Feed Back encryption mode. Is allowed only if GRD_SOFTWARE_ACCELERATION flag is provided
        GRD_NO_COUNTER_DECREMENT  = 0x100,      // Execute encryption/decryption without counter decrement (used another cryptographic key)
        GRD_SOFTWARE_ACCELERATION = 0x200       // Software acceleration for encryption/decryption of a large amount of data
    }

    public enum GrdDriverType : int
    {
        GRD_DRIVER_TYPE_NONE    = 0,
        GRD_DRIVER_TYPE_HID     = 1,
        GRD_DRIVER_TYPE_WINUSB  = 2,
        GRD_DRIVER_TYPE_VENDOR  = 3
    }

    [Flags]
    public enum GrdFirmwareState : int
    {
        GRD_FIRMWARE_STATE_OK = 0,
        GRD_FIRMWARE_STATE_CRC_ERROR = 1,
        GRD_REMOTE_UPDATE_MODE_ON = 2,
        GRD_FIRMWARE_STATE_TIMER_FREQUENCY_ERROR = 4,
        GRD_FIRMWARE_STATE_TIMER_INTEGRITY_ERROR = 8,
        GRD_FIRMWARE_STATE_TIMER_BATTERY_ERROR = 16
    }

    public enum GrdMemoryWriteMode
    {
        GRD_WRITE_MODE_MOV = 0x00,       // Destination becomes source
        GRD_WRITE_MODE_XOR = 0x01        // Destination becomes source XORed with destination
    }

    public enum GrdLanguageId : int
    {
        GRD_LANG_EN = 0,
        GRD_LANG_RU = 1,
    }

    public class VendorCodes
    {
        public UInt32 m_publicCode;
        public UInt32 m_privateReadCode;
        public UInt32 m_privateWriteCode;

        public VendorCodes(UInt32 publicCode, UInt32 privateReadCode = 0, UInt32 privateWriteCode = 0)
        {
            this.m_publicCode = publicCode;
            this.m_privateReadCode = privateReadCode;
            this.m_privateWriteCode = privateWriteCode;
        }
    }

    public class CustomerInfo
    {
        public string m_firstName;
        public string m_lastName;
        public string m_email;
        public string m_companyName;
        public string m_phone;

        public CustomerInfo(string firstName, string lastName, string email, string companyName, string phone)
        {
            m_firstName = firstName;
            m_lastName = lastName;
            m_email = email;
            m_companyName = companyName;
            m_phone = phone;
        }
    }

    public class GrdAesContext
    {
        public byte[] iv = new byte[16];  // initialization vector
        public byte[] inf = new byte[8];  // additional encryption info
    };

    public class GrdProductFeaturePair
    {
        public UInt32 productNumber;
        public UInt32 featureNumber;

        public GrdProductFeaturePair(UInt32 productNumber, UInt32 featureNumber)
        {
            this.productNumber = productNumber;
            this.featureNumber = featureNumber;
        }
    };
}
