#include <string.h>
#include <stdint.h>
#include <assert.h>
#include <jni.h>
#include "log.h"
#include "jniFunction.h"

jfieldID GetFieldID(JNIEnv* env, jobject obj, const char* name, const char* sig)
{
    assert(env);
    assert(sig);
    assert(name);

    if (env == NULL)
    {
        LOGE("Error! JNIEnv* is NULL/n");
        return NULL;
    }

    if (name == NULL)
    {
        LOGE("Error! Field Name is NULL/n");
        return NULL;
    }

    if (sig == NULL)
    {
        LOGE("Error! Signature of field %s is NULL/n", name);
        return NULL;
    }

    if (obj == NULL)
    {
        LOGE("Error! Jobject %s is NULL/n", name);
        return NULL;
    }

    jclass cls = env->GetObjectClass(obj);
    if (cls == NULL)
    {
        LOGE("Error! GetObjectClass Name = %s Signature = %s", name, sig);
        return NULL;
    }

    jfieldID fieldID = env->GetFieldID(cls, name, sig);
    if (fieldID == NULL)
        LOGE("Error! GetFieldID Name = %s Signature = %s", name, sig);

    env->DeleteLocalRef(cls);
    return fieldID;
}

bool SetByteField(JNIEnv* env, jobject obj, const char* name, jbyte value)
{
    jfieldID fieldID = GetFieldID(env, obj, name, "B");
    if (fieldID == NULL)
        return false;

    env->SetByteField(obj, fieldID, value);
    return true;
}

bool SetShortField(JNIEnv* env, jobject obj, const char* name, jshort value)
{
    jfieldID fieldID = GetFieldID(env, obj, name, "S");
    if (fieldID == NULL)
        return false;

    env->SetShortField(obj, fieldID, value);
    return true;
}

bool SetIntField(JNIEnv* env, jobject obj, const char* name, jint value)
{
    jfieldID fieldID = GetFieldID(env, obj, name, "I");
    if (fieldID == NULL)
        return false;

    env->SetIntField(obj, fieldID, value);
    return true;
}


bool SetLongField(JNIEnv* env, jobject obj, const char* name, jlong value)
{
    jfieldID fieldID = GetFieldID(env, obj, name, "J");
    if (fieldID == NULL)
        return false;

    env->SetLongField(obj, fieldID, value);
    return true;
}

bool SetObjectField(JNIEnv* env, jobject obj, const char* name, const char* sig, jobject value)
{
    jfieldID fieldID = GetFieldID(env, obj, name, sig);
    if (fieldID == NULL)
        return false;

    env->SetObjectField(obj, fieldID, value);
    return true;
}

bool GetByteField(JNIEnv* env, jobject obj, const char* name, jbyte* value)
{
    jfieldID fieldID = GetFieldID(env, obj, name, "B");
    if (fieldID == NULL)
        return false;

    *value = env->GetByteField(obj, fieldID);
    return true;
}

bool GetShortField(JNIEnv* env, jobject obj, const char* name, jshort* value)
{
    jfieldID fieldID = GetFieldID(env, obj, name, "S");
    if (fieldID == NULL)
        return false;

    *value = env->GetShortField(obj, fieldID);
    return true;
}

bool GetIntField(JNIEnv* env, jobject obj, const char* name, jint* value)
{
    jfieldID fieldID = GetFieldID(env, obj, name, "I");
    if (fieldID == NULL)
        return false;

    *value = env->GetIntField(obj, fieldID);
    return true;
}

bool GetLongField(JNIEnv* env, jobject obj, const char* name, jlong* value)
{
    jfieldID fieldID = GetFieldID(env, obj, name, "J");
    if (fieldID == NULL)
        return false;

    *value = env->GetLongField(obj, fieldID);
    return true;
}



bool GetStaticLongField(JNIEnv* env, const char* cls_name, const char* name, jlong* ret_value)
{
    jclass cls = env->FindClass(cls_name);
    if (cls == NULL)
        return false;

    jfieldID fieldID = env->GetStaticFieldID(cls, name, "J");
    if (fieldID == NULL)
    {
        LOGE("Error! GetStaticFieldID Name = %s Signature = %s", name, "J");
        return false;
    }

    *ret_value = env->GetStaticLongField(cls, fieldID);
    return true;
}



bool GetObjectField(JNIEnv* env, jobject obj, const char* name, const char* sig, jobject* value)
{
    jfieldID fieldID = GetFieldID(env, obj, name, sig);
    if (fieldID == NULL)
        return false;

    *value = env->GetObjectField(obj, fieldID);
    return true;
}


jmethodID GetMethodID(JNIEnv* env, jobject obj, const char* method_name, const char* method_sig)
{
    assert(env);
    assert(method_name);
    assert(method_sig);

    if (env == NULL)
    {
        LOGE("Error! JNIEnv* is NULL/n");
        return NULL;
    }

    if (method_name == NULL)
    {
        LOGE("Error! Method Name is NULL/n");
        return NULL;
    }

    if (method_sig == NULL)
    {
        LOGE("Error! Signature of method %s is NULL/n", method_name);
        return NULL;
    }

    if (obj == NULL)
    {
        LOGE("Error! Object for method %s is NULL/n", method_name);
        return NULL;
    }

    jclass cls = env->GetObjectClass(obj);
    if (cls == NULL)
    {
        LOGE("Error! GetObjectClass for method %s /n", method_name);
        return NULL;
    }

    jmethodID mid = env->GetMethodID(cls, method_name, method_sig);
    if (mid == NULL)
    {
        LOGE("Error! GetMethodID %s from class /n", method_name);
        return NULL;
    }

    env->DeleteLocalRef(cls);
    return mid;
}

jmethodID GetStaticMethodID(JNIEnv* env, const char* cls_name, const char* method_name, const char* method_sig, jclass* cls)
{
    *cls = env->FindClass(cls_name);
    if (*cls == NULL)
        return NULL;

    return env->GetStaticMethodID(*cls, method_name, method_sig);
}


bool CallStaticIntMethod(JNIEnv* env, const char* cls_name, const char* method_name, const char* method_sig, jint* ret_value, ...)
{
    va_list args;
    va_start(args, ret_value);

    jclass cls;
    jmethodID mid = GetStaticMethodID(env, cls_name, method_name, method_sig, &cls);
    if (mid == NULL)
    {
        va_end(args);
        return false;
    }

    *ret_value = env->CallStaticIntMethod(cls, mid);
    env->DeleteLocalRef(cls);

    va_end(args);
    return true;
}


bool CallStaticObjectMethod(JNIEnv* env, const char* cls_name, const char* method_name, const char* method_sig, jobject* ret_value, ...)
{
    va_list args;
    va_start(args, ret_value);

    jclass cls;
    jmethodID mid = GetStaticMethodID(env, cls_name, method_name, method_sig, &cls);
    if (mid == NULL)
    {
        va_end(args);
        return false;
    }

    *ret_value = env->CallStaticObjectMethod(cls, mid);
    env->DeleteLocalRef(cls);

    va_end(args);
    return true;
}

bool CallVoidMethodV(JNIEnv* env, jobject obj, const char* method_name, const char* method_sig, ...)
{
    va_list args;
    va_start(args, method_sig);

    jmethodID mid = GetMethodID(env, obj, method_name, method_sig);
    if (mid == NULL)
    {
        va_end(args);
        return false;
    }

    env->CallVoidMethodV(obj, mid, args);

    va_end(args);
    return true;
}

bool CallIntMethodV(JNIEnv* env, jobject obj, const char* method_name, const char* method_sig, jint* ret_value, ...)
{
    va_list args;
    va_start(args, ret_value);

    jmethodID mid = GetMethodID(env, obj, method_name, method_sig);
    if (mid == NULL)
    {
        va_end(args);
        return false;
    }

    *ret_value = env->CallIntMethodV(obj, mid, args);

    va_end(args);
    return true;
}



bool CallObjectMethodV(JNIEnv* env, jobject obj, const char* method_name, const char* method_sig, jobject* ret_value, ...)
{
    va_list args;
    va_start(args, ret_value);

    jmethodID mid = GetMethodID(env, obj, method_name, method_sig);
    if (mid == NULL)
    {
        va_end(args);
        return false;
    }

    *ret_value = env->CallObjectMethodV(obj, mid, args);
    va_end(args);
    return true;
}

bool CallBooleanMethodV(JNIEnv* env, jobject obj, const char* method_name, const char* method_sig, jboolean* ret_value, ...)
{
    va_list args;
    va_start(args, ret_value);

    jmethodID mid = GetMethodID(env, obj, method_name, method_sig);
    if (mid == NULL)
    {
        va_end(args);
        return false;
    }

    *ret_value = env->CallBooleanMethodV(obj, mid, args);
    va_end(args);
    return true;
}


bool CreateObjectV(JNIEnv* env, const char* cls_name, const char* ctor_name, const char* ctor_sign, jobject* newObj, ...)
{
    va_list args;
    va_start(args, newObj);

    jclass cls = env->FindClass(cls_name);
    if (NULL == cls)
    {
        LOGE("Error! Not found class name %s is NULL/n", cls_name);
        va_end(args);
        return false;
    }

    // Get the Method ID of the constructor which takes an int
    jmethodID midCtor = env->GetMethodID(cls, ctor_name, ctor_sign);
    if (NULL == midCtor)
    {
        LOGE("Error! Signature of ctor method %s is NULL/n", ctor_name);
        va_end(args);
        return false;
    }

    // Call back constructor to allocate a new instance, with an int argument
    *newObj = env->NewObjectV(cls, midCtor, args);
    if (NULL == *newObj)
    {
        LOGE("Error! Create object/n");
        va_end(args);
        return false;
    }

    va_end(args);
    return true;

}


