#! python3

# Notes:
# - several functions not implemented here
#   - GrdSetDriverType
#   - GrdMemoryGetSize, GrdMemoryRead, GrdMemoryWrite
# - prototypes can be used for Grd* functions (with ctypes _FuncPtr restype
#     and argtypes)

from ctypes import cdll, c_uint, c_char, c_char_p, c_void_p, string_at, byref, Structure
from ctypes.util import find_library
from enum import Enum, IntEnum, IntFlag
from os import path
from sys import platform as _platform

_GRDLIC_SEARCH_PATH = "./" # can be modified by the importing module via ResetGrdlicSearchPath()
grdlic = None

def ResetGrdlicSearchPath(path):
    global _GRDLIC_SEARCH_PATH
    _GRDLIC_SEARCH_PATH = path
    retval = _Load()
    if not retval:
        print("Unable to find the grdlic dynamic library!")
    return retval

def _GetName():
    if _platform == "win32" or _platform == "win64":
        return "grdlic.dll"
    elif _platform == "linux" or _platform == "linux2":
        return "libgrdlic.so"
    elif _platform == "darwin":
        return "grdlic.dylib"
    else:
        assert False
        return "grdlic"

def _GetLibraryName():
    if _platform == "win32" or _platform == "win64":
        return "grdlic.dll"
    elif _platform == "linux" or _platform == "linux2":
        return "grdlic"
    elif _platform == "darwin":
        return "grdlic.dylib"
    else:
        assert False
        return "grdlic"

def _Load():

    global grdlic

    try:
        grdlic = cdll.LoadLibrary(path.join(_GRDLIC_SEARCH_PATH, _GetName()))
    except FileNotFoundError as e: # Windows
        #print(e)
        grdlic = None
    except OSError as e: # Linux
        #print(e)
        grdlic = None

    if grdlic != None:
        return True

    search = find_library(_GetLibraryName())
    if search == None:
        search = str()

    try:
        grdlic = cdll.LoadLibrary(search)
    except FileNotFoundError as e: # Windows
        #print(e)
        grdlic = None
    except OSError as e: # Linux
        #print(e)
        grdlic = None

    return grdlic != None

_Load()

grd_uint32 = c_uint
grd_uint8 = c_char

class Status(Enum):
    OK                                          = 0          # Successful operation status
    INVALID_HANDLE                              = 1          # Invalid handle
    FEATURE_NOT_FOUND                           = 2          # Feature not found
    FEATURE_RESOURCE_EXHAUST                    = 3          # Feature network resource exhausted
    INVALID_FEATURE                             = 4          # Invalid feature data
    INVALID_SERIAL_NUMBER                       = 5          # Serial number is invalid
    INVALID_LICENSE                             = 6          # Can't read DL license file because of hardware mismatched or part of file header is corrupted
    LICENSE_NOT_ACTIVATED                       = 7          # License with this licenseId has not been activated yet or reactivation required
    LICENSE_ALREADY_ACTIVATED                   = 8          # License already activated
    GET_HARDWARE_ID_ERROR                       = 9          # Can not get hardware id
    NOT_ENOUGH_HARWARE_TO_BIND                  = 10         # Insufficient number of computer components to activate
    REQUEST_FAILED                              = 12         # Unable to send request
    GET_RESPONSE_FAILED                         = 13         # Unable to receive response
    SERVER_NOT_FOUND                            = 14         # Unable to find server
    INCORRECT_RESPONSE_DATA                     = 15         # Incorrect response format or data not valid
    FILE_CREATION_ERROR                         = 16         # Error during creation of file (e.g. license file)
    SERIAL_NUMBER_RESOURCE_EXHAUSTED            = 17         # Serial number resource exhausted
    NO_UPDATES_AVAILABLE                        = 18         # No updates available for license
    ACTIVATION_NOT_AVAILABLE                    = 19         # Activation not enabled by software vendor reasons
    NO_CUSTOMER_SPECIFIED                       = 20         # No customer info specified
    CAN_NOT_ACTIVATE_NON_DEMO_LICENSE           = 21         # Can't activate non demo license
    OUTDATED_API_VERSION                        = 22         # API version too low
    INVALID_PARAMETER                           = 23         # Invalid function argument
    BUFFER_TOO_SMALL                            = 24         # Buffer is not enough to perform operation
    INTERNAL_ERROR                              = 25         # Internal error
    LICENSE_IS_BANNED                           = 26         # Software vendor has banned serial number and license
    NO_LICENSE_AVAILABLE                        = 27         # No license available (e.g. order access time is over)
    FEATURE_RUNCOUNTER_EXHAUST                  = 28         # Feature execution counter is exhausted
    FEATURE_EXPIRED                             = 29         # Feature life time is expired
    INVALID_PUBLIC_KEY                          = 30         # Invalid public key
    NO_SERVICE                                  = 31         # Service is not supported
    UNABLE_SEND_REQUEST_TO_ADMIN_RUNTIME        = 32         # Admin runtime is not installed or outdated
    DONGLE_NOT_FOUND                            = 33         # Guardant dongle not found
    NO_RESULTS_FOUND                            = 34         # No any results found
    INVALID_DIGEST                              = 35         # Invalid message or digest
    FEATURE_INACTIVE                            = 36         # The period of using feature has not yet arrived or the feature is not active
    MEMORY_OUT_OF_RANGE                         = 37         # Access to memory beyond its range
    ACCESS_DENIED                               = 38         # Access denied
    NUMBER_ATTEMPTS_EXHAUSTED                   = 39         # The number of password attempts has been exhausted
    SERIAL_NUMBER_NOT_FOUND                     = 40         # Such serial number not found on server
    GRD_INVALID_MEMORY                          = 41         # Invalid memory ID specified or item does not exist
    VIRTUAL_ENVIRONMENT_DETECTED                = 42         # Running in virtual environment detected
    REMOTE_DESKTOP_DETECTED                     = 43         # Detected running in remote desktop mode
    OUTDATED_GCC_VERSION                        = 44         # Outdated Gurdant Control Center version
    REHOST_IS_NOT_ALLOWED                       = 45         # Rehost is not allowed
    STORAGE_CORRUPTED                           = 46         # DL storage corrupted
    STORAGE_ID_MISMATCH                         = 47         # DL storage ID mismatch (input is not intended for current storage)
    INVALID_UPDATE_OR_ALREADY_INSTALLED         = 48         # Invalid update for DL received or update has been installed before
    INVALID_REHOST_DATA                         = 49         # Invalid rehost date sent to server
    HARDWARE_ID_MISMATCH                        = 50         # Hardware id mismatch
    REACTIVATION_IS_NOT_ALLOWED                 = 51         # Reactivation is not allowed because of license contains write memory or features with counter
    RECIPIENT_HAS_THE_SAME_LICENSE              = 52         # Recipient has the same license activated as the license trying to rehost to it
    CAN_NOT_DETACH_SOME_FEATURES                = 53         # Detach failed for some of chosen features
    CANCEL_PREVIOUS_DETACH_REQUIRED             = 54         # Required cancel previous detach before making new one
    DETACHED_LICENSE_EXPIRED                    = 55         # Detached license already exprired
    DETACHED_LICENSE_NOT_FOUND                  = 56         # Detached license not found
    DETACH_FORBIDDEN_BY_GCC                     = 57         # Detach forbidden by Gurdant Control Center
    TRIAL_LICENSE_EXPIRED                       = 58         # Trial license expired
    ORERATION_IS_NOT_ALLOWED                    = 59         # Operation is not allowed
    RTC_ERROR                                   = 60         # Timer error
    FUNCTION_NOT_FOUND                          = 0x70000000 # Function not found
    DLL_NOT_FOUND                               = 0x70000001 # DLL not found
    MANAGE_ERROR                                = 0x70000002 # Unknown error

    def __str__(self):
        return self.name # avoid "Status." prefix

class RemoteMode(IntFlag):
    LOCAL  = 0x01
    REMOTE = 0x02

class DongleModel(IntFlag):
    SIGN = 0x80
    CODE = 0x100
    DL   = 0x400

class NetServer(Structure):

    _HOST_NAME_LEN = 256
    _BROADCAST_ADDRESS_LEN = 256
    _RESERVED_LEN = 0x200

    _fields_ = [("HostName", c_char * _HOST_NAME_LEN),
                ("HostPort", grd_uint32),
                ("BroadcastAddress", c_char * _BROADCAST_ADDRESS_LEN),
                ("Reserved", grd_uint8 * _RESERVED_LEN)]

    def __init__(self, hostName, hostPort, broadcastAddress):
        if hostName != None:
            self.HostName = hostName
        self.HostPort = hostPort
        if broadcastAddress != None:
            self.BroadcastAddress = broadcastAddress

class VendorCodes(Structure):

    _fields_ = [("PublicCode", grd_uint32),
                ("PrivateReadCode", grd_uint32),
                ("PrivateWriteCode", grd_uint32)]

    def __init__(self, publicCode, privateReadCode, privateWriteCode):
        self.PublicCode = publicCode
        self.PrivateReadCode = privateReadCode
        self.PrivateWriteCode = privateWriteCode

class FeatureEncryptMode(IntFlag):
    GRD_EM_ECB                = 0x00  # Electronic Code Book encryption mode (set by default). Data size must be multiple of 16
    GRD_EM_CBC                = 0x01  # Cipher Block Chaining encryption mode. Data size must be multiple of 16
    GRD_EM_CFB                = 0x02  # Cipher Feed Back encryption mode. Is allowed only if GRD_SOFTWARE_ACCELERATION flag is provided
    GRD_EM_OFB                = 0x04  # Output Feed Back encryption mode. Is allowed only if GRD_SOFTWARE_ACCELERATION flag is provided
    GRD_NO_COUNTER_DECREMENT  = 0x100 # Execute encryption/decryption without counter decrement (used another cryptographic key)
    GRD_SOFTWARE_ACCELERATION = 0x200 # Software acceleration for encryption/decryption of a large amount of data

class GrdAesContext(Structure):

    _IV_SIZE = 16
    _INF_SIZE = 8

    _pack_ = 1

    _fields_ = [("iv", grd_uint8 * _IV_SIZE), # initialization vector
                ("inf", grd_uint8 * _INF_SIZE)] # additional encryption info

    def __init__(self, iv, inf):
        self.iv = iv
        self.inf = inf

class GrdECC160(IntEnum):
    PUBLIC_KEY_SIZE = 40
    PRIVATE_KEY_SIZE = 20
    DIGEST_SIZE = 40
    MESSAGE_SIZE = 20

class CustomerInfo(Structure):

    _FIRST_NAME_LEN = 64
    _LAST_NAME_LEN = 64
    _EMAIL_LEN = 128
    _COMPANY_NAME_LEN = 128
    _PHONE_LEN = 32

    _pack_ = 1

    _fields_ = [("FirstName", c_char * _FIRST_NAME_LEN),
                ("LastName", c_char * _LAST_NAME_LEN),
                ("Email", c_char * _EMAIL_LEN),
                ("CompanyName", c_char * _COMPANY_NAME_LEN),
                ("Phone", c_char * _PHONE_LEN)]

class GrdLanguageId(IntEnum):
    GRD_LANG_EN = 0
    GRD_LANG_RU = 1

# Standalone functions

def GetApiVersion():

    # example: version 2.3 -> major == 2, minor == 3, coded == 0x00020003

    majorVersion = grd_uint32()
    minorVersion = grd_uint32()
    codedVersion = grd_uint32(grdlic.GrdGetApiVersion(byref(majorVersion), byref(minorVersion)))
    return [majorVersion.value, minorVersion.value, codedVersion.value]

def GetLicenseInfo(visibilityStr, vendorCodes):

    # returns the JSON-formatted string with descriptions for all corresponding licenses

    vendorCodesBytes = c_char_p()
    if vendorCodes != None:
        vendorCodesBytes = c_char_p(bytes(vendorCodes))

    licenseInfoJsonArrayPtr = c_char_p()
    visibilityBytes = c_char_p(visibilityStr.encode('utf-8'))
    status = Status(grd_uint32(grdlic.GrdGetLicenseInfo(visibilityBytes, vendorCodesBytes, byref(licenseInfoJsonArrayPtr))).value)
    licenseInfoJsonArray = licenseInfoJsonArrayPtr.value
    if licenseInfoJsonArrayPtr != c_char_p():
        grdlic.GrdFree(licenseInfoJsonArrayPtr)
    return [status, licenseInfoJsonArray]

def GetSessionInfo(visibilityStr, vendorCodes):

    # returns the JSON-formatted string with descriptions for all corresponding sessions

    vendorCodesBytes = c_char_p()
    if vendorCodes != None:
        vendorCodesBytes = c_char_p(bytes(vendorCodes))

    sessionInfoJsonArrayPtr = c_char_p()
    visibilityBytes = c_char_p(visibilityStr.encode('utf-8'))
    status = Status(grd_uint32(grdlic.GrdGetSessionInfo(visibilityBytes, vendorCodesBytes, byref(sessionInfoJsonArrayPtr))).value)
    sessionInfoJsonArray = sessionInfoJsonArrayPtr.value
    if sessionInfoJsonArrayPtr != c_char_p():
        grdlic.GrdFree(sessionInfoJsonArrayPtr)
    return [status, sessionInfoJsonArray]

def GetHostInfo(visibilityStr):

    # returns the JSON-formatted string with descriptions for all detected hosts

    hostInfoJsonArrayPtr = c_char_p()
    visibilityBytes = c_char_p(visibilityStr.encode('utf-8'))
    status = Status(grd_uint32(grdlic.GrdGetHostInfo(visibilityBytes, byref(hostInfoJsonArrayPtr))).value)
    hostInfoJsonArray = hostInfoJsonArrayPtr.value
    if hostInfoJsonArrayPtr != c_char_p():
        grdlic.GrdFree(hostInfoJsonArrayPtr)
    return [status, hostInfoJsonArray]

def VerifyDigest(publicKey, message, digest):

    assert len(publicKey) == GrdECC160.PUBLIC_KEY_SIZE
    assert len(digest) == GrdECC160.DIGEST_SIZE
    status = Status(grd_uint32(grdlic.GrdVerifyDigest(grd_uint32(len(publicKey)), c_char_p(publicKey), grd_uint32(len(message)), c_char_p(message), grd_uint32(len(digest)), c_char_p(digest))).value)
    return status

def CheckLicenseSerialNumberFormat(serialNumber):

    serialNumberBytes = c_char_p(serialNumber.encode('utf-8'))
    status = Status(grd_uint32(grdlic.GrdLicenseCheckSerialNumberFormat(serialNumberBytes)).value)
    return status

def LedBlink(dongleId):

    status = Status(grd_uint32(grdlic.GrdLedBlink(grd_uint32(dongleId))).value)
    return status

def GetErrorMessage(statusCode, languageId):

    assert languageId == GrdLanguageId.GRD_LANG_EN or languageId == GrdLanguageId.GRD_LANG_RU

    status = Status.DLL_NOT_FOUND # random non-OK code
    size = 64 # random limit
    while True:

        message = bytes(size)
        status = Status(grd_uint32(grdlic.GrdGetErrorMessage(Status(statusCode).value, languageId, c_char_p(message), grd_uint32(len(message)))).value)
        if status != Status.INVALID_PARAMETER:
            break
        size = size * 2

    return [status, message.decode('utf-8').rstrip("\0")]

INVALID_LICENSE_ID_VALUE = 0xffffffff # constant
INVALID_FEATURE_NUMBER_VALUE = 0xffffffff # constant

class Feature:

    INVALID_FEATURE_HANDLE_VALUE = 0xffffffff # constant

    def __init__(self, featureNumber):

        self.number = grd_uint32(featureNumber)
        self.handle = grd_uint32(Feature.INVALID_FEATURE_HANDLE_VALUE)

    def Login(self, vendorCodes, visibilityStr):

        vendorCodesBytes = c_char_p()
        if vendorCodes != None:
            vendorCodesBytes = c_char_p(bytes(vendorCodes))

        visibilityBytes = c_char_p(visibilityStr.encode('utf-8'))
        status = Status(grd_uint32(grdlic.GrdFeatureLogin(self.number, vendorCodesBytes, visibilityBytes, byref(self.handle))).value)
        return status

    def Logout(self):

        status = Status(grd_uint32(grdlic.GrdFeatureLogout(self.handle)).value)
        self.handle = grd_uint32(Feature.INVALID_FEATURE_HANDLE_VALUE)
        return status

    def __del__(self):

        if (self.handle.value != Feature.INVALID_FEATURE_HANDLE_VALUE):
            self.Logout()

    def GetInfo(self):

        licenseInfoPtr = c_char_p()
        status = Status(grd_uint32(grdlic.GrdFeatureGetInfo(self.handle, byref(licenseInfoPtr))).value)
        licenseInfo = licenseInfoPtr.value
        if licenseInfoPtr != c_char_p():
            grdlic.GrdFree(licenseInfoPtr)
        return [status, licenseInfo]

    def Check(self, publicKey):

        publicKeySize = grd_uint32()
        publicKeyBytes = c_char_p()
        if publicKey != None:
            assert len(publicKey) == GrdECC160.PUBLIC_KEY_SIZE
            publicKeySize = grd_uint32(len(publicKey))
            publicKeyBytes = c_char_p(publicKey)
        status = Status(grd_uint32(grdlic.GrdFeatureCheck(self.handle, publicKeySize, publicKeyBytes)).value)
        return status

    def Encrypt(self, data, mode, context):

        # no checks for data size or flags consistency is performed here, this
        #   job is delegated to the grdlic API
        # ECB and CBC demand data size be multiple of 16
        # ECB does not need context, other modes (CBC, CFB, OBF) do
        # CFB and OFB demand software acceleration flag be set
        # flag for no counter decrement uses another encryption key, hence
        #   encryption with decrement and decryption without decrement cannot
        #   result in the original data

        if mode == FeatureEncryptMode.GRD_EM_ECB or context == None:
            contextSize = grd_uint32()
            contextBytes = c_char_p()
        else:
            contextSize = 24
            contextBytes = c_char_p(bytes(context))

        status = Status(grd_uint32(grdlic.GrdFeatureEncrypt(self.handle, grd_uint32(len(data)), c_char_p(data), mode, contextBytes, contextSize)).value)
        return status

    def Decrypt(self, data, mode, context):

        # no checks for data size or flags consistency is performed here, this
        #   job is delegated to the grdlic API
        # ECB and CBC demand data size be multiple of 16
        # ECB does not need context, other modes (CBC, CFB, OBF) do
        # CFB and OFB demand software acceleration flag be set
        # flag for no counter decrement uses another encryption key, hence
        #   encryption with decrement and decryption without decrement cannot
        #   result in the original data

        if mode == FeatureEncryptMode.GRD_EM_ECB or context == None:
            contextSize = grd_uint32()
            contextBytes = c_char_p()
        else:
            contextSize = 24
            contextBytes = c_char_p(bytes(context))

        status = Status(grd_uint32(grdlic.GrdFeatureDecrypt(self.handle, grd_uint32(len(data)), c_char_p(data), mode, contextBytes, contextSize)).value)
        return status

    def Sign(self, message, digest):

        assert len(digest) == GrdECC160.DIGEST_SIZE
        status = Status(grd_uint32(grdlic.GrdFeatureSign(self.handle, grd_uint32(len(message)), c_char_p(message), grd_uint32(len(digest)), c_char_p(digest))).value)
        return status

    def GetTimeLimit(self):

        remainingTime = grd_uint32()
        status = Status(grd_uint32(grdlic.GrdFeatureGetTimeLimit(self.handle, byref(remainingTime))).value)
        return [status, remainingTime.value]

    def GetRunCounter(self):

        runCounter = grd_uint32()
        status = Status(grd_uint32(grdlic.GrdFeatureGetRunCounter(self.handle, byref(runCounter))).value)
        return [status, runCounter.value]

    def GetRealTime(self):

        currentTime = grd_uint32()
        status = Status(grd_uint32(grdlic.GrdGetRealTime(self.handle, byref(currentTime))).value)
        return [status, currentTime.value]

    def GetMaxConcurrentResource(self):

        maxConcurrentResource = grd_uint32()
        status = Status(grd_uint32(grdlic.GrdFeatureGetMaxConcurrentResource(self.handle, byref(maxConcurrentResource))).value)
        return [status, maxConcurrentResource.value]

class LicenseActivator:

    def __init__(self, host, port):

        self.host = host
        self.port = port

    def Activate(self, serialNumber, customerInfo):

        serialNumberBytes = c_char_p(serialNumber.encode('utf-8'))
        hostBytes = c_char_p(self.host.encode('utf-8'))
        licenseId = grd_uint32()
        status = Status(grd_uint32(grdlic.GrdLicenseActivate(serialNumberBytes, hostBytes, grd_uint32(self.port), c_char_p(bytes(customerInfo)), byref(licenseId))).value)
        return [status, licenseId.value]

    def Update(self, licenseId):

        hostBytes = c_char_p(self.host.encode('utf-8'))
        status = Status(grd_uint32(grdlic.GrdLicenseUpdate(grd_uint32(licenseId), hostBytes, grd_uint32(self.port))).value)
        return status

    def CheckUpdateIsAvailable(self, licenseId):

        hostBytes = c_char_p(self.host.encode('utf-8'))
        status = Status(grd_uint32(grdlic.GrdLicenseCheckUpdateIsAvailable(grd_uint32(licenseId), hostBytes, grd_uint32(self.port))).value)
        return status

    def CheckIsNotBanned(self, licenseId):

        hostBytes = c_char_p(self.host.encode('utf-8'))
        status = Status(grd_uint32(grdlic.GrdLicenseCheckIsNotBanned(grd_uint32(licenseId), hostBytes, grd_uint32(self.port))).value)
        return status

    def Remove(self, licenseId):

        hostBytes = c_char_p(self.host.encode('utf-8'))
        status = Status(grd_uint32(grdlic.GrdLicenseRemove(grd_uint32(licenseId))).value)
        return status

    def CreateActivationRequest(self):

        requestBytes = c_char_p()
        requestSize = grd_uint32()
        status = Status(grd_uint32(grdlic.GrdLicenseCreateActivationRequest(byref(requestBytes), byref(requestSize))).value)
        assert len(requestBytes.value) == requestSize.value - 1
        activationRequest = requestBytes.value
        if requestBytes != c_char_p():
            grdlic.GrdFree(requestBytes)
        return [status, activationRequest]

    def SendActivationRequest(self, serialNumber, customerInfo, activationRequest):

        serialNumberBytes = c_char_p(serialNumber.encode('utf-8'))
        requestBytes = c_char_p(activationRequest)
        requestSize = len(requestBytes.value) + 1
        hostBytes = c_char_p(self.host.encode('utf-8'))
        responseBytes = c_char_p()
        responseSize = grd_uint32()
        status = Status(grd_uint32(grdlic.GrdLicenseSendActivationRequest(serialNumberBytes, c_char_p(bytes(customerInfo)), requestBytes, requestSize, hostBytes, grd_uint32(self.port), byref(responseBytes), byref(responseSize))).value)
        activationResponse = string_at(responseBytes, size=responseSize.value)
        if responseBytes != c_char_p():
            grdlic.GrdFree(responseBytes)
        return [status, activationResponse]

    def CreateUpdateRequest(self, licenseId):

        requestBytes = c_char_p()
        requestSize = grd_uint32()
        status = Status(grd_uint32(grdlic.GrdLicenseCreateUpdateRequest(grd_uint32(licenseId), byref(requestBytes), byref(requestSize))).value)
        assert len(requestBytes.value) == requestSize.value - 1
        updateRequest = requestBytes.value
        if requestBytes != c_char_p():
            grdlic.GrdFree(requestBytes)
        return [status, updateRequest]

    def SendUpdateRequest(self, updateRequest):

        requestBytes = c_char_p(updateRequest)
        requestSize = len(requestBytes.value) + 1
        hostBytes = c_char_p(self.host.encode('utf-8'))
        responseBytes = c_char_p()
        responseSize = grd_uint32()
        status = Status(grd_uint32(grdlic.GrdLicenseSendUpdateRequest(requestBytes, requestSize, hostBytes, grd_uint32(self.port), byref(responseBytes), byref(responseSize))).value)
        updateResponse = string_at(responseBytes, size=responseSize.value)
        if responseBytes != c_char_p():
            grdlic.GrdFree(responseBytes)
        return [status, updateResponse]

    def Install(self, serverResponse):

        status = Status(grd_uint32(grdlic.GrdLicenseInstall(c_char_p(serverResponse), grd_uint32(len(serverResponse)))).value)
        return status
