//------------------------------------------------------------------------------
// (C) 1997 - 2019 Aktiv Co. All rights reserved.
//  Guardant Licensing API example C++
//------------------------------------------------------------------------------
// This example demonstrates Guardant Licensing API usage.
// This code sample is for API demonstration only!
// Please do not use this code in real projects!
//------------------------------------------------------------------------------
#include <cstring>
#include <vector>
#include <iostream>
#include <sstream>
#include <time.h>
#ifdef WIN32
#include <windows.h>
#endif
#include <grdlic/grdlic_cpp.h>

// demo access codes
static const grd_uint32 DEMO_ACCESSCODE_PUBLIC = 0x519175b7Lu;        // Demo public code
static const grd_uint32 DEMO_ACCESSCODE_PRIVATEREAD = 0x51917645Lu;   // Demo private read code
static const grd_uint32 DEMO_ACCESSCODE_PRIVATEWRITE = 0x51917603Lu;  // Demo private write code

void PrintError(const std::string& Text);
void PrintText(const std::string& Text);
void PrintSuccess();
int PrintCode(int nErrorCode);
int ErrorHandling(int nRet, int exitProgram);

//----------------------------------------------------------------------------------
//  Main function
//----------------------------------------------------------------------------------
int main(int argc, char** argv)
{
    try
    {
        grd_uint32 featureId = 0;
        guardant::VendorCodes vendorCodes(DEMO_ACCESSCODE_PUBLIC, DEMO_ACCESSCODE_PRIVATEREAD, 0);

        //------------------------------------------------------------------------------
        // Setup license and feature search params
        //------------------------------------------------------------------------------
        std::string visibility = "{\"dongleModel\": 0x480, \"remoteMode\": 3}";

        //------------------------------------------------------------------------------
        // Retrieve information about all available licenses
        //------------------------------------------------------------------------------
        {
            std::string licenseInfoList;
            std::cout << "Retrieve information about all available licenses: ";
            guardant::GetLicenseInfo(visibility, NULL, licenseInfoList);
            if (licenseInfoList.empty())
            {
                PrintError("No licenses found\n");
                exit(-1);
            }

            std::cout << "Information about all available licenses:\n" << licenseInfoList.c_str() << std::endl;

            // Set up feature id from command line (if specified) or set arbitrary
            if (argc > 1)
                featureId = atoi(argv[1]);
            else
                featureId = 1;

            std::cout << "For further work chosen feature with ID: " << featureId << std::endl;
        }

        //------------------------------------------------------------------------------
        // Login to feature
        //------------------------------------------------------------------------------
        guardant::Feature feature(featureId, vendorCodes, visibility);
        std::cout << "Login to the feature with id " << featureId;
        feature.Login();
        PrintSuccess();

        //------------------------------------------------------------------------------
        // Get information about the feature we are logged in
        //------------------------------------------------------------------------------
        {
            std::string outLicenseInfo;
            std::cout << "Get information about the feature we are logged in: ";
            feature.GetInfo(outLicenseInfo);
            PrintSuccess();

            std::cout << "Information about the feature we are logged in:\n" << outLicenseInfo.c_str() << std::endl;
        }

        //------------------------------------------------------------------------------
        // Encryption and decryption by AES128
        //------------------------------------------------------------------------------
        {
            size_t bufSize = 128;
            grd_uint8 buf[128];
            grd_uint8 outBuf[128];
            memset(buf, 0xFF, bufSize);

            std::cout << "Encrypt some data using the feature with id " << featureId << " and taking into account"
                " it's license limitations (run counter or life time period): ";
            memcpy(outBuf, buf, bufSize);
            feature.Encrypt(bufSize, outBuf, GrdFeatureEncryptMode::GRD_EM_ECB, NULL, 0);
            PrintSuccess();
            std::cout << "Decrypt previously encrypted data using the feature with id " << featureId << " and taking into account"
                " it's license limitations (run counter or life time period): ";
            feature.Decrypt(bufSize, outBuf, GrdFeatureEncryptMode::GRD_EM_ECB, NULL, 0);
            PrintSuccess();

            if (memcmp(buf, outBuf, bufSize) != 0)
            {
                PrintError("Result of decryption mismatched with origin data\n!");
                exit(-1);
            }

            std::cout << "Encrypt some data using the feature with id " << featureId << " without licensing limitations: ";
            memcpy(outBuf, buf, bufSize);
            feature.Encrypt(bufSize, outBuf,
                            static_cast<GrdFeatureEncryptMode>(GrdFeatureEncryptMode::GRD_NO_COUNTER_DECREMENT | GrdFeatureEncryptMode::GRD_EM_ECB),
                            NULL, 0);
            PrintSuccess();
            std::cout << "Decrypte previously encrypted data using the feature with id " << featureId << " without licensing limitations: ";
            feature.Decrypt(bufSize, outBuf,
                            static_cast<GrdFeatureEncryptMode>(GrdFeatureEncryptMode::GRD_NO_COUNTER_DECREMENT | GrdFeatureEncryptMode::GRD_EM_ECB),
                            NULL, 0);
            PrintSuccess();

            if (memcmp(buf, outBuf, bufSize) != 0)
            {
                PrintError("Result of decryption mismatched with origin data\n!");
                exit(-1);
            }
        }

        //------------------------------------------------------------------------------
        // Digitally sign a block of bytes
        //------------------------------------------------------------------------------
        {
            char digest[GRD_ECC160_DIGEST_SIZE];
            char message[GRD_ECC160_MESSAGE_SIZE] = "__Message for sign_";
            grd_uint8 publicKey[] = "\x1D\xCE\x42\x91\xEF\xF6\x8C\x6A\xEC\xB6\xC6\x76\x7E\xF0\xCC\x0B\x61\xD5\xA1\x73\x69\x14\x53\x05\xC3\xFD\x98\x93\xD6\x0A\xF8\xFA\x03\xE1\x8C\x37\x99\x4B\x95\x6B";

            std::cout << "Digitally sign a message: ";
            feature.Sign(GRD_ECC160_MESSAGE_SIZE, message, GRD_ECC160_DIGEST_SIZE, digest);
            PrintSuccess();

            //        std::cout << "Verifing digitally sign a message: ";
            //        feature.VerifySign(GRD_ECC160_PUBLIC_KEY_SIZE, publicKey, GRD_ECC160_MESSAGE_SIZE, message, GRD_ECC160_DIGEST_SIZE, digest);
            //        PrintSuccess();
        }

        //------------------------------------------------------------------------------
        // Get rest of the feature life time (valid for time limited feature only)
        //------------------------------------------------------------------------------
        try {
            grd_time_t remainingTime = 0;
            std::cout << "Get rest of the feature life time: ";
            feature.GetTimeLimit(remainingTime);
            PrintSuccess();
            std::cout << "Rest of feature life time in seconds: " << remainingTime << std::endl;
        } catch (guardant::Exception& e)
        {
            ErrorHandling(e.status(), 0);
        }

        //------------------------------------------------------------------------------
        // Get current date and time (valid for time limited feature only)
        //------------------------------------------------------------------------------
        try {
            grd_time_t currentTime = 0;
            feature.GetRealTime(currentTime);
            PrintSuccess();
            std::cout << "Current datetime in seconds: " << currentTime << std::endl;
        } catch (guardant::Exception& e)
        {
            ErrorHandling(e.status(), 0);
        }

        //------------------------------------------------------------------------------
        // Working with memory (valid for feature with memory only)
        //------------------------------------------------------------------------------
        if (argc > 2)
        {
            grd_uint32 memoryId = atoi(argv[2]);
            std::cout << "Get memory size: ";
            grd_uint32 memSize;
            try {
                feature.MemoryGetSize(memoryId, memSize);
                PrintSuccess();
            } catch (guardant::Exception& e)
            {
                ErrorHandling(e.status(), 0);
                memSize = 4; // 4 bytes by default
            }

            std::cout << "Memory size: " << memSize << std::endl;

            std::cout << "Read memory: ";
            grd_uint32 password = 0; // no password
            grd_uint32 offset = 0;
            grd_uint8* memBuf = (grd_uint8*)malloc((size_t)memSize);
            try {
                feature.MemoryRead(memoryId, password, memSize, offset, memBuf);
                free(memBuf);
                PrintSuccess();
            } catch (guardant::Exception& e)
            {
                free(memBuf);
                ErrorHandling(e.status(), 0);
            }
        }
    } // try
    catch (guardant::Exception& e)
    {
        ErrorHandling(e.status(), 1);
    }
    catch (...)
    {
        PrintError("Unknown error detected!\n");
        exit(-1);
    }

    PrintText("\nALL SUCCESSFULLY DONE!\n");

    return 0;
}

//----------------------------------------------------------------------------------------------
// Prints operation result and forces program termination on error
//----------------------------------------------------------------------------------------------
void PrintError(const std::string& Text)
{
#ifdef WIN32
    SetConsoleTextAttribute(GetStdHandle(STD_OUTPUT_HANDLE), FOREGROUND_RED);
    std::cout << Text << std::endl;
    SetConsoleTextAttribute(GetStdHandle(STD_OUTPUT_HANDLE), FOREGROUND_GREEN | FOREGROUND_RED | FOREGROUND_BLUE);
#elif defined(__linux) || defined(__linux__)
    std::cout << "\033[1;31m";
    std::cout << Text << std::endl;
    std::cout << "\033[0m";
#else
    std::cout << Text << std::endl;
#endif
}

//----------------------------------------------------------------------------------------------
// Prints operation result and forces program termination on error
//----------------------------------------------------------------------------------------------
void PrintText(const std::string& Text)
{
#ifdef WIN32
    SetConsoleTextAttribute(GetStdHandle(STD_OUTPUT_HANDLE), FOREGROUND_GREEN);
    std::cout << Text << std::endl;
    SetConsoleTextAttribute(GetStdHandle(STD_OUTPUT_HANDLE), FOREGROUND_GREEN | FOREGROUND_RED | FOREGROUND_BLUE);
#elif defined(__linux) || defined(__linux__)
    std::cout << "\033[1;32m";
    std::cout << Text << std::endl;
    std::cout << "\033[0m";
#else
    std::cout << Text << std::endl;
#endif
}

//----------------------------------------------------------------------------------------------
// Prints operation result and forces program termination on error
//----------------------------------------------------------------------------------------------
int PrintCode(int nErrorCode)
{
    // Buffer for error string
    if (nErrorCode == GRD_OK)
        PrintText("success");
    else
    {
        std::ostringstream ss;
        ss << "ERROR! " << nErrorCode << " ";
        PrintError(ss.str());
    }
    return nErrorCode;
}

//----------------------------------------------------------------------------------------------
// Handle errors
// Prints operation result, logout from feature
//----------------------------------------------------------------------------------------------
int ErrorHandling(int nRet, int exitProgram)
{
    PrintCode(nRet);

    if (nRet != GRD_OK && exitProgram)
    {
        // Terminate application
        PrintError("Application terminated by exit()");
        exit(nRet);
    }
    return nRet;
}
//------------------------------------------------------------------------------

//----------------------------------------------------------------------------------------------
// Prints success word
//------------------------------------------------------------------------------
void PrintSuccess()
{
    PrintText("success");
}
//------------------------------------------------------------------------------

