using System;
using System.Collections;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.Security.Cryptography;
using System.Xml;
using System.Runtime.Serialization.Json;
using System.Xml.Linq;
using grdlic;

namespace grdlic.test
{
    /// <summary>
    /// C# example for using Grdlic API
    /// </summary>
    class Program
    {
        static void HandleError(Status s)
        {
            Console.WriteLine("Status: " + s.ToString());
            if (s != Status.OK)
            {
                Console.WriteLine("Error. Quit program");
                Environment.Exit(-1);
            }
        }

        static void PrintLicenseInfo(string licenseInfo, out UInt32 firstFeatureNumber)
        {
            firstFeatureNumber = 0;

            var x = new XmlDictionaryReaderQuotas();
            var jsonReader = JsonReaderWriterFactory.CreateJsonReader(Encoding.UTF8.GetBytes(licenseInfo), x);
            var license = XElement.Load(jsonReader);

            PrintLicenseInfo(license, out firstFeatureNumber);
        }

        static void PrintLicenseInfo(XElement license, out UInt32 firstFeatureNumber)
        {
            firstFeatureNumber = 0;

            int isBroken = Convert.ToInt32(license.Element("isBroken").Value);
            if (isBroken == 1)
            {
                Console.WriteLine("License {0} in DL is broken", license.Attribute("licenseId").Value);
                return;
            }

            var licenseInfo = license.Element("licenseInfo");
            if (licenseInfo == null || licenseInfo.IsEmpty)
            {
                var dongleInfo = license.Element("dongleInfo");
                Console.WriteLine("Hardware dongle without license. Dongle: ID {0}. Model: {1}. Vendor public code: {2}",
                                                                 dongleInfo.Element("dongleId").Value,
                                                                 dongleInfo.Element("dongleModel").Value,
                                                                 dongleInfo.Element("publicCode").Value);
                return;
            }

            Console.WriteLine("License {0} vendor public {1} vendor company {2} products count {3} ",
                                                              licenseInfo.Element("licenseId").Value,
                                                              licenseInfo.Element("vendorPublicCode").Value,
                                                              licenseInfo.Element("vendorCompanyName").Value,
                                                              licenseInfo.Element("productsCount").Value);

            var products = licenseInfo.Elements("products");
            Console.WriteLine("\tProducts:");
            foreach (XElement product in products.Elements())
            {
                Console.WriteLine("\tProduct {0} with number {1} [modification {2}] with {3} features",
                                  product.Element("name").Value,
                                  product.Element("number").Value,
                                  product.Element("modification").Value,
                                  product.Element("featuresCount").Value);

                var features = product.Elements("features");
                Console.WriteLine("\t\tFeatures:");
                bool setFirstFeatureNumber = false;
                foreach (XElement feature in features.Elements())
                {
                    if (!setFirstFeatureNumber)
                    {
                        firstFeatureNumber = Convert.ToUInt32(feature.Element("number").Value);
                        setFirstFeatureNumber = true;
                    }

                    Console.WriteLine("\t\tFeature {0} with number {1}",
                                      feature.Element("name").Value,
                                      feature.Element("number").Value,
                                      feature.Element("restOfLifeTime").Value
                                      );
                }
            }
        }

        static void PrintLicensesInfo(string licenses, out UInt32 firstFeatureNumber)
        {
            firstFeatureNumber = 0;

            if (licenses == null)
                return;

            var x = new XmlDictionaryReaderQuotas();
            var jsonReader = JsonReaderWriterFactory.CreateJsonReader(Encoding.UTF8.GetBytes(licenses), x);
            var root = XElement.Load(jsonReader);

            var lic = root.Elements("licenses");

            if (lic == null)
                return;

            foreach (XElement el in lic.Elements())
            {
                PrintLicenseInfo(el, out firstFeatureNumber);
            }
        }

        static int Main(string[] args)
        {
            Console.WriteLine("\nExample of using Grdlic API");
            try
            {
                // call GrdGetApiVersion
                Console.WriteLine("Call GrdGetApiVersion");
                UInt32 v, major, minor;
                Status r = GrdlicApi.GrdGetApiVersion(out v, out major, out minor);
                HandleError(r);
                Console.WriteLine("API version: " + major.ToString() + "." + minor.ToString());

                // get info about available licenses across this PC (because of m_remoteMode = RemoteMode.LOCAL)
                string visibility = "{\"dongleModel\": 0x480, \"remoteMode\": 3}";
                Console.WriteLine("Call GrdGetLicenseInfo");
                string licenses;
                r = GrdlicApi.GrdGetLicenseInfo(visibility, null, out licenses);
                HandleError(r);

                Console.WriteLine("ALL AVAILABLE LICENSES:");
                // get some feature to work with
                UInt32 featureNumber;
                PrintLicensesInfo(licenses, out featureNumber);

                // do login to feature
                Console.WriteLine("Try login feature {0}", featureNumber);
                GrdlicApi.Feature feature = new GrdlicApi.Feature(featureNumber);
                VendorCodes vendorAccessCodes = new VendorCodes(0x519175b7, 0x51917645);
                r = feature.Login(vendorAccessCodes, visibility);
                HandleError(r);

                // get feature license info
                Console.WriteLine("Get feature {0} info", featureNumber);
                string licenseInfo;
                r = feature.GetInfo(out licenseInfo);
                HandleError(r);
                // print feature info
                Console.WriteLine("FEATURE INFO:");
                UInt32 dummy;
                PrintLicenseInfo(licenseInfo, out dummy);

                // check feature
                Console.WriteLine("Check feature");
                r = feature.Check(null);
                HandleError(r);

                // encrypt data
                Console.WriteLine("Encrypt data");
                byte[] data = {1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16};
                SHA1 shaAlgo = new SHA1CryptoServiceProvider();
                byte[] hash = shaAlgo.ComputeHash(data);
                r = feature.Encrypt(data, FeatureEncryptMode.GRD_EM_ECB, null);
                HandleError(r);
                // decrypt data
                Console.WriteLine("Decrypt data");
                r = feature.Decrypt(data, FeatureEncryptMode.GRD_EM_ECB, null);
                HandleError(r);
                bool compareResult = StructuralComparisons.StructuralEqualityComparer.Equals(hash, shaAlgo.ComputeHash(data));
                Console.WriteLine("Decryption sanity check: {0}", compareResult);
                if (!compareResult)
                    Environment.Exit(-1);

                // unlimited encrypt data
                Console.WriteLine("Encrypt data");
                hash = shaAlgo.ComputeHash(data);
                r = feature.Encrypt(data, FeatureEncryptMode.GRD_NO_COUNTER_DECREMENT | FeatureEncryptMode.GRD_EM_ECB, null);
                HandleError(r);
                // unlimited decrypt data
                Console.WriteLine("Decrypt data");
                r = feature.Decrypt(data, FeatureEncryptMode.GRD_NO_COUNTER_DECREMENT | FeatureEncryptMode.GRD_EM_ECB, null);
                HandleError(r);
                compareResult = StructuralComparisons.StructuralEqualityComparer.Equals(hash, shaAlgo.ComputeHash(data));
                Console.WriteLine("Unlimited decryption sanity check: {0}", compareResult);
                if (!compareResult)
                    Environment.Exit(-1);

                // sign data
                Console.WriteLine("Sign data");
                byte[] message = new byte[(int)GrdECC160.MESSAGE_SIZE];
                byte[] digest;
                r = feature.Sign(message, out digest);
                HandleError(r);
//                // verify sign [! required feature correct public key]
//                Console.WriteLine("Verify sign");
//                byte[] publicKey = {0x1D, 0xCE, 0x42, 0x91, 0xEF, 0xF6, 0x8C, 0x6A, 0xEC, 0xB6, 0xC6, 0x76, 0x7E, 0xF0, 0xCC, 0x0B, 0x61, 0xD5, 0xA1, 0x73, 0x69, 0x14, 0x53, 0x05, 0xC3, 0xFD, 0x98, 0x93, 0xD6, 0x0A, 0xF8, 0xFA, 0x03, 0xE1, 0x8C, 0x37, 0x99, 0x4B, 0x95, 0x6B};
//                r = GrdlicApi.GrdVerifySign(publicKey, message, digest);
//                HandleError(r);

                // get feature rest of the lifetime in seconds
                Console.WriteLine("Get feature rest of the lifetime");
                long remainingTime;
                r = feature.GetTimeLimit(out remainingTime);
                if (r == Status.NO_SERVICE)
                    Console.WriteLine("NO SERVICE (skip)");
                else
                    HandleError(r);
                Console.WriteLine("Remaining time: {0}", remainingTime);

                // get rest of feature run counter value
                Console.WriteLine("Get feature rest of the run counter value");
                UInt32 runCounter;
                r = feature.GetRunCounter(out runCounter);
                if (r == Status.NO_SERVICE)
                    Console.WriteLine("NO SERVICE (skip)");
                else
                    HandleError(r);
                Console.WriteLine("Run counter value: {0}", runCounter);

                // get time from real clock in seconds
                Console.WriteLine("Get time from real clock");
                long currentTime;
                r = feature.GetRealTime(out currentTime);
                HandleError(r);
                Console.WriteLine("Current time: {0}", currentTime);

                // get feature max concurrent resource value
                Console.WriteLine("Get feature max concurrent resource value");
                UInt32 value;
                r = feature.GetMaxConcurrentResource(out value);
                HandleError(r);
                Console.WriteLine("Feature max concurrent resource value: {0}", value);

                // working with memory (valid for feature with memory only)
                if (args.Length > 0)
                {
                    UInt32 memoryId = Convert.ToUInt32(args[1]);
                    Console.WriteLine("Get memory size");
                    long memSize;
                    r = feature.MemoryGetSize(memoryId, out memSize);
                    if (r != Status.OK)
                    {
                        // use default
                        memSize = 4; // 4 bytes by default
                    }

                    Console.WriteLine("Memory size: {0}", memSize);

                    Console.WriteLine("Read memory");
                    UInt32 password = 0; // no password
                    UInt32 offset = 0;
                    byte[] memBuf = new byte[memSize];
                    r = feature.MemoryRead(memoryId, password, offset, memBuf);
                    HandleError(r);
                }

                // logout from feature
                Console.WriteLine("Logout from feature {0}", featureNumber);
                r = feature.Logout();
                HandleError(r);
            } catch(Exception e)
            {
                Console.WriteLine(e.ToString());
            }

            return 0;
        }
    }
}
