//------------------------------------------------------------------------------
// (C) 1997 - 2019 Aktiv Co. All rights reserved.
//  Guardant License acitvation API example C++
//------------------------------------------------------------------------------
// This example demonstrates Guardant Licensing API usage.
// This code sample is for API demonstration only!
// Please do not use this code in real projects!
//------------------------------------------------------------------------------
#include <cstring>
#include <vector>
#include <iostream>
#include <memory>
#include <sstream>
#include <cstdlib>
#include <time.h>
#ifdef WIN32
#include <windows.h>
#endif
#include <grdlic/grdlic_cpp.h>

void PrintError(const std::string& Text);
void PrintText(const std::string& Text);
void PrintSuccess();
int PrintCode(int nErrorCode);
int ErrorHandling(int nRet);

template<typename ... Args>
std::string string_format(const std::string& format, Args ... args)
{
    size_t size = snprintf(nullptr, 0, format.c_str(), args ...) + 1;   // Extra space for '\0'
    std::unique_ptr<char[]> buf(new char[size]);
    std::snprintf(buf.get(), size, format.c_str(), args ...);
    return std::string(buf.get(), buf.get() + size - 1);   // We don't want the '\0' inside
}

//----------------------------------------------------------------------------------
//  Main function
//----------------------------------------------------------------------------------
int main(int argc, char** argv)
{
    try
    {
        if (argc < 4)
        {
            printf("Usage:"
                   "\tdemo <serial> <host> <port>\n");
            exit(-1);
        }

        //------------------------------------------------------------------------------
        // Set customer info
        //------------------------------------------------------------------------------
        GrdCustomerInfo customerInfo = {0};
        strcpy(customerInfo.companyName, "demo");
        strcpy(customerInfo.email, "demo@customer.ru");
        strcpy(customerInfo.phone, "+7(777)777-77-77");

        //------------------------------------------------------------------------------
        // Activate License
        //------------------------------------------------------------------------------
        printf("Activate license: ");
        grd_uint32 licenseId = 0;
        std::string serialNumber = argv[1];
        guardant::LicenseActivator activator(argv[2], std::atoi(argv[3]));
        activator.Activate(serialNumber, &customerInfo, licenseId);
        PrintSuccess();

        //------------------------------------------------------------------------------
        // Get license info
        //------------------------------------------------------------------------------
        {
            std::string visibilityFmt = "{\"licenseId\": %d, \"dongleModel\": 0x400, \"remoteMode\": 1}";
            std::string visibility = string_format(visibilityFmt, licenseId);
            std::string licenseInfoList;
            printf("Retrieve information about license: ");
            guardant::GetLicenseInfo(visibility, NULL, licenseInfoList);
            if (licenseInfoList.empty())
            {
                PrintError("No licenses found\n");
                exit(-1);
            }

            printf("Information about license:\n%s\n", licenseInfoList.c_str());
        }

        //------------------------------------------------------------------------------
        // Get license info after activation
        //------------------------------------------------------------------------------
        printf("Check license update after initial activation: ");
        try {
            activator.CheckUpdateIsAvailable(licenseId);
        } catch (guardant::Exception& e) {
            if (e.status() == GRD_NO_UPDATES_AVAILABLE)
            {
                PrintText("no updates available. \n\nSUCCESSFULLY DONE!\n");
                exit(0);
            }
            else if (e.status() != GRD_OK)
                throw;
            else
                PrintSuccess(); // updates available
        }

        //------------------------------------------------------------------------------
        // License update
        //------------------------------------------------------------------------------
        printf("Update license: ");
        activator.Update(licenseId);
        PrintSuccess();

        //------------------------------------------------------------------------------
        // Get license info after license update
        //------------------------------------------------------------------------------
        {
            std::string visibilityFmt = "{\"licenseId\": %d, \"dongleModel\": 0x400, \"remoteMode\": 1}";
            std::string visibility = string_format(visibilityFmt, licenseId);
            std::string licenseInfoList;
            printf("Retrieve information about license: ");
            guardant::GetLicenseInfo(visibility, NULL, licenseInfoList);
            if (licenseInfoList.empty())
            {
                PrintError("No licenses found\n");
                exit(-1);
            }

            printf("Information about license:\n%s\n", licenseInfoList.c_str());
        }

        //------------------------------------------------------------------------------
        // Check license update one more time
        //------------------------------------------------------------------------------
        printf("Check license update after update [must return GRD_NO_UPDATES_AVAILABLE]: ");
        try {
            activator.CheckUpdateIsAvailable(licenseId);
        } catch (guardant::Exception& e) {
            if (e.status() == GRD_NO_UPDATES_AVAILABLE)
            {
                PrintText("no updates available. \n\nSUCCESSFULLY DONE!\n");
                exit(0);
            }
            else if (e.status() != GRD_OK)
                throw;
        }

        PrintText("\n\nWARNING: updates still available!\n");

        return 1;
    } // try
    catch (guardant::Exception& e)
    {
        ErrorHandling(e.status());
        exit(-1);
    }
    catch (...)
    {
        PrintError("Unknown error detected!\n");
        exit(-1);
    }

    PrintText("\nALL SUCCESSFULLY DONE!\n");

    return 0;
}

//----------------------------------------------------------------------------------------------
// Prints operation result and forces program termination on error
//----------------------------------------------------------------------------------------------
void PrintError(const std::string& Text)
{
#ifdef WIN32
    SetConsoleTextAttribute(GetStdHandle(STD_OUTPUT_HANDLE), FOREGROUND_RED);
    std::cout << Text << std::endl;
    SetConsoleTextAttribute(GetStdHandle(STD_OUTPUT_HANDLE), FOREGROUND_GREEN | FOREGROUND_RED | FOREGROUND_BLUE);
#elif defined(__linux) || defined(__linux__)
    std::cout << "\033[1;31m";
    std::cout << Text << std::endl;
    std::cout << "\033[0m";
#else
    std::cout << Text << std::endl;
#endif
}

//----------------------------------------------------------------------------------------------
// Prints operation result and forces program termination on error
//----------------------------------------------------------------------------------------------
void PrintText(const std::string& Text)
{
#ifdef WIN32
    SetConsoleTextAttribute(GetStdHandle(STD_OUTPUT_HANDLE), FOREGROUND_GREEN);
    std::cout << Text << std::endl;
    SetConsoleTextAttribute(GetStdHandle(STD_OUTPUT_HANDLE), FOREGROUND_GREEN | FOREGROUND_RED | FOREGROUND_BLUE);
#elif defined(__linux) || defined(__linux__)
    std::cout << "\033[1;32m";
    std::cout << Text << std::endl;
    std::cout << "\033[0m";
#else
    std::cout << Text << std::endl;
#endif
}

//----------------------------------------------------------------------------------------------
// Prints operation result and forces program termination on error
//----------------------------------------------------------------------------------------------
int PrintCode(int nErrorCode)
{
    // Buffer for error string
    if (nErrorCode == GRD_OK)
        PrintText("success");
    else
    {
        std::ostringstream ss;
        ss << "ERROR! " << nErrorCode << " ";
        PrintError(ss.str());
    }
    return nErrorCode;
}

//----------------------------------------------------------------------------------------------
// Handle errors
// Prints operation result, logout from feature
//----------------------------------------------------------------------------------------------
int ErrorHandling(int nRet)
{
    PrintCode(nRet);

    if (nRet != GRD_OK)
    {
        // Terminate application
        PrintError("Application terminated by exit()");
        exit(nRet);
    }
    return nRet;
}
//------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------------
// Prints success word
//------------------------------------------------------------------------------
void PrintSuccess()
{
    PrintText("success");
}
//------------------------------------------------------------------------------

