using System;
using System.Collections;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.Security.Cryptography;
using System.Xml;
using System.Runtime.Serialization.Json;
using System.Xml.Linq;
using grdlic;

namespace grdlic.test
{
    /// <summary>
    /// C# example for using Grdlic API for license activation
    /// </summary>
    class Program
    {
        static void HandleError(Status s)
        {
            Console.WriteLine("Status: " + s.ToString());

            if (s == Status.NO_UPDATES_AVAILABLE)
                return;

            if (s != Status.OK)
            {
                Console.WriteLine("Error. Quit program");
                Environment.Exit(-1);
            }
        }

        static bool getFirstLicenseId(string licenses, out UInt32 licenseId)
        {
            licenseId = 0;

            if (licenses == null)
                return false;

            var x = new XmlDictionaryReaderQuotas();
            var jsonReader = JsonReaderWriterFactory.CreateJsonReader(Encoding.UTF8.GetBytes(licenses), x);
            var root = XElement.Load(jsonReader);

            var lic = root.Elements("licenses");

            if (lic == null)
                return false;

            foreach (XElement el in lic.Elements())
            {
                int isBroken = Convert.ToInt32(el.Element("isBroken").Value);
                if (isBroken == 1)
                {
                    Console.WriteLine("License {0} in DL is broken", el.Element("licenseId").Value);
                    continue;
                }

                var licenseInfo = el.Element("licenseInfo");
                if (licenseInfo == null || licenseInfo.IsEmpty)
                {
                    var dongleInfo = el.Element("dongleInfo");
                    Console.WriteLine("Hardware dongle without license. Dongle: ID {0}. Model: {1}. Vendor public code: {2}",
                                                                     dongleInfo.Element("dongleId").Value,
                                                                     dongleInfo.Element("dongleModel").Value,
                                                                     dongleInfo.Element("publicCode").Value);
                    continue;
                }

                licenseId = (UInt32)Convert.ToInt64(licenseInfo.Element("licenseId").Value);
                return true;
            }

            return false;
        }

        static int Main(string[] args)
        {
            Console.WriteLine("\nExample of using Grdlic API for license activation");
            try
            {
                if (args == null || args.Length == 0)
                {
                    Console.WriteLine("Serial number not specifed");
                    return 1;
                }

                Status r = Status.OK;

                // activate license
                Console.WriteLine("Activate license by serial: {0}", args[0]);
                GrdlicApi.LicenseActivator la = new GrdlicApi.LicenseActivator("getlicense.guardant.ru", 443);
                CustomerInfo customer = new CustomerInfo("Ivan", "Ivanov", "ba@go.ru", "", "");
                UInt32 licenseId;
                r = la.Activate(args[0], customer, out licenseId);
                HandleError(r);
                Console.WriteLine("Activated license ID: {0}", licenseId);

                // get info about available licenses across this PC (because of m_remoteMode = RemoteMode.LOCAL)
                string visibility = "{\"dongleModel\": 0x480, \"remoteMode\": 3}";
                Console.WriteLine("Call GrdGetLicenseInfo");
                string licenses;
                r = GrdlicApi.GrdGetLicenseInfo(visibility, null, out licenses);
                HandleError(r);

                Console.WriteLine("ALL AVAILABLE LICENSES:");
                Console.WriteLine(licenses);

                if (!getFirstLicenseId(licenses, out licenseId))
                {
                    Console.WriteLine("No licenses detected.\nError. Quit program");
                    Environment.Exit(-1);
                }

                // update license
                Console.WriteLine("Update license with ID: {0}", licenseId);
                r = la.Update(licenseId);
                HandleError(r);

                // check license update is available
                Console.WriteLine("Check license update is available");
                r = la.CheckUpdateIsAvailable(licenseId);
                HandleError(r);

                // check license is not banned
                Console.WriteLine("Check license is not banned");
                r = la.CheckIsNotBanned(licenseId);
                HandleError(r);

                // remove license
                Console.WriteLine("Remove license");
                r = la.Remove(licenseId);
                HandleError(r);

                // OFFLINE ACTIVATION API
                // create license activation request
                Console.WriteLine("Create license activation request");
                byte[] activationRequest;
                r = la.CreateActivationRequest(out activationRequest);
                HandleError(r);

                // send activation request
                Console.WriteLine("Send license activation request");
                byte[] activationResponse;
                r = la.SendActivationRequest(args[0], customer, activationRequest, out activationResponse);
                HandleError(r);

                // install license
                Console.WriteLine("Install license");
                r = GrdlicApi.InstallLicense(activationResponse);
                HandleError(r);

                // get info about available licenses across this PC
                Console.WriteLine("Call GrdGetLicenseInfo");
                r = GrdlicApi.GrdGetLicenseInfo(visibility, null, out licenses);
                HandleError(r);

                Console.WriteLine("ALL AVAILABLE LICENSES:");
                Console.WriteLine(licenses);

                if (!getFirstLicenseId(licenses, out licenseId))
                {
                    Console.WriteLine("No licenses detected.\nError. Quit program");
                    Environment.Exit(-1);
                }

                // create license update request
                Console.WriteLine("Create license update request");
                byte[] updateRequest;
                r = la.CreateUpdateRequest(licenseId, out updateRequest);
                HandleError(r);

                // send update request
                Console.WriteLine("Send license update request");
                byte[] updateResponse;
                r = la.SendUpdateRequest(updateRequest, out updateResponse);
                HandleError(r);

                if (r == Status.OK)
                {
                    // install updated license
                    Console.WriteLine("Install updated license");
                    r = GrdlicApi.InstallLicense(updateResponse);
                    HandleError(r);
                }

                // remove license
                Console.WriteLine("Remove license");
                r = la.Remove(licenseId);
                HandleError(r);
            } catch(Exception e)
            {
                Console.WriteLine(e.ToString());
            }

            return 0;
        }
    }
}
