#! python3

# Note that feature public key can be obtained from the Guardant Station
#   web browser interface, it is marked as for usage in the GrdVerifyDigest
#   function

from sys import exit, argv
from json import loads
from hashlib import sha256
from datetime import datetime, timedelta

from grdlic_python import Status, LicenseActivator, CustomerInfo, GetLicenseInfo, GetErrorMessage, CheckLicenseSerialNumberFormat, LedBlink, INVALID_LICENSE_ID_VALUE, INVALID_FEATURE_NUMBER_VALUE, RemoteMode, DongleModel, GrdLanguageId

def uhex(value):
    return hex(value & (2**32 - 1)) # 32 for int size

def handle_error(status):

    # print as enum
    print("Status: " + str(status))

    ## print corresponding error message
    #[status_, message] = GetErrorMessage(status, GrdLanguageId.GRD_LANG_EN)
    #print(message)

    if (status != Status.OK and status != Status.NO_SERVICE and status != Status.LICENSE_ALREADY_ACTIVATED and status != Status.NO_UPDATES_AVAILABLE):
        print("Error! Program exits.")
        exit(-1)
    return

def print_license_info(license):

    licenseId = INVALID_LICENSE_ID_VALUE
    lastFeatureNumber = INVALID_FEATURE_NUMBER_VALUE

    isBroken = license["isBroken"]
    if (isBroken == 1):
        print("  License " + uhex(license["licenseId"]) + " in DL is broken!")
        return [licenseId, lastFeatureNumber]

    try:
        licenseInfo = license["licenseInfo"]
    except KeyError:
        dongleInfo = license["dongleInfo"]
        print("  Hardware dongle without license:\n" +
              "    Dongle ID: " + uhex(dongleInfo["dongleId"]) + "\n" +
              "    Dongle model: " + uhex(dongleInfo["dongleModel"]) + "\n" +
              "    Vendor public code: " + uhex(dongleInfo["publicCode"]))

        LedBlink(dongleInfo["dongleId"])
        #status = LedBlink(dongleInfo["dongleId"])
        #handle_error(status)

        return [licenseId, lastFeatureNumber]

    print("  License " + uhex(licenseInfo["licenseId"]) + " is OK:\n" +
          "    Vendor public code: " + uhex(licenseInfo["vendorPublicCode"]) + "\n" +
          "    Vendor company name: " + licenseInfo["vendorCompanyName"] + "\n" +
          "    Products count: " + str(licenseInfo["productsCount"]))

    licenseId = licenseInfo["licenseId"]

    products = licenseInfo["products"]

    for product in products:

        print("      Product \"" + product["name"] + "\"" +
                   " with number " + str(product["number"]) +
                   " (modification " + str(product["modification"]) + ")" +
                   " with " + str(product["featuresCount"]) + " features")

        features = product["features"]

        for feature in features:

            print("        Feature \"" + feature["name"] + "\"" +
                         " with number " + str(feature["number"]))

            lastFeatureNumber = feature["number"]

    return [licenseId, lastFeatureNumber]

def print_licenses_info(licenseInfoJsonArray):

    root = loads(licenseInfoJsonArray)

    lastLicenseId = INVALID_LICENSE_ID_VALUE
    lastFeatureNumber = INVALID_FEATURE_NUMBER_VALUE

    if root["licenseCount"] == 0:
        print("  No licenses found!")
        return [lastLicenseId, lastFeatureNumber]

    licenses = root["licenses"]
    for license in licenses:
        [lastLicenseId, lastFeatureNumber] = print_license_info(license)

    return [lastLicenseId, lastFeatureNumber]

def test_activate_online(serialNumber, la, customerInfo):

    print("\nOnline activation\n")

    # activate license
    print("Call GrdLicenseActivate")
    [status, licenseId] = la.Activate(serialNumber, customerInfo)
    handle_error(status)
    print("Activated license with ID " + uhex(licenseId))

    # get info about all licenses available at this PC (because of m_remoteMode = RemoteMode.LOCAL)
    visibilityStr = "{\"dongleModel\": 0x480, \"remoteMode\": 3}"
    print("Call GrdGetLicenseInfo")
    [status, licenseInfoJsonArray] = GetLicenseInfo(visibilityStr, None) # vendor access codes not specified
    handle_error(status)

    # print info about all licenses available at this PC
    print("ALL AVAILABLE LICENSES:")
    [licenseId, lastFeatureNumber] = print_licenses_info(licenseInfoJsonArray)

    if lastFeatureNumber == INVALID_FEATURE_NUMBER_VALUE:
        return

    # check if license update is available
    print("Call GrdLicenseCheckUpdateIsAvailable")
    status = la.CheckUpdateIsAvailable(licenseId)
    handle_error(status)

    # update license
    print("Call GrdLicenseUpdate")
    status = la.Update(licenseId)
    handle_error(status)

    # check if license is not banned
    print("Call GrdLicenseCheckIsNotBanned")
    status = la.CheckIsNotBanned(licenseId)
    handle_error(status)

    # remove license
    print("Call GrdLicenseRemove")
    status = la.Remove(licenseId)
    handle_error(status)

def test_activate_offline(serialNumber, la, customerInfo):

    print("\nOffline activation\n")

    # create license activation request
    print("Call GrdLicenseCreateActivationRequest")
    [status, activationRequest] = la.CreateActivationRequest()
    handle_error(status)

    # send license activation request
    print("Call GrdLicenseSendActivationRequest")
    [status, activationResponse] = la.SendActivationRequest(serialNumber, customerInfo, activationRequest)
    handle_error(status)

    # install license
    print("Call GrdLicenseInstall")
    status = la.Install(activationResponse)
    handle_error(status)

    # get info about all licenses available at this PC (because of m_remoteMode = RemoteMode.LOCAL)
    visibilityStr = "{\"dongleModel\": 0x480, \"remoteMode\": 3}"
    print("Call GrdGetLicenseInfo")
    [status, licenseInfoJsonArray] = GetLicenseInfo(visibilityStr, None) # vendor access codes not specified
    handle_error(status)

    # print info about all licenses available at this PC
    print("ALL AVAILABLE LICENSES:")
    [licenseId, lastFeatureNumber] = print_licenses_info(licenseInfoJsonArray)

    if lastFeatureNumber == INVALID_FEATURE_NUMBER_VALUE:
        return

    # create license update request
    print("Call GrdLicenseCreateUpdateRequest")
    [status, updateRequest] = la.CreateUpdateRequest(licenseId)
    handle_error(status)

    # send license update request
    print("Call GrdLicenseSendUpdateRequest")
    [status, updateResponse] = la.SendUpdateRequest(updateRequest)
    handle_error(status)

    if status != Status.NO_UPDATES_AVAILABLE:
        print("Call GrdLicenseInstall")
        status = la.Install(updateResponse)
        handle_error(status)

    # removing the license
    print("Call GrdLicenseRemove")
    status = la.Remove(licenseId)
    handle_error(status)

def activate(serialNumber):

    print("Activating license by serial number: " + serialNumber)
    la = LicenseActivator("getlicense.guardant.ru", 443) # default is non-standalone Guardant Station
    customerInfo = CustomerInfo(b"Ivan", b"Ivanov", b"ba@go.ru", b"", b"")

    # check license serial number format
    print("Call GrdLicenseCheckSerialNumberFormat")
    status = CheckLicenseSerialNumberFormat(serialNumber)
    handle_error(status)

    test_activate_online(serialNumber, la, customerInfo)
    test_activate_offline(serialNumber, la, customerInfo)

    return

def test_activation(serials):

    print("\n--------------------------------------------------\n"
            "Example of using grdlic API for license activation"
          "\n--------------------------------------------------\n")

    if serials == []:
        print("No serial numbers specified!")
        return

    for serial in serials:
        activate(serial)

    return

def main(argv):
    test_activation(argv)

if __name__ == "__main__":
    main(argv[1:])
